@@ -24,7 +24,11 @@ from datetime import date, datetime
 import multiprocessing as mp
 import pyproj
 from time import sleep
+import dill
 
+# dill.Pickler.dumps, dill.Pickler.loads = dill.dumps, dill.loads
+# mp.reduction.ForkingPickler = dill.Pickler
+# mp.reduction.dump = dill.dump
 
 ###############################################
 ### Parameters
@@ -453,32 +457,34 @@ def combine_metadata(project, dataset):
     return datasets
 
 
-def multi_calc_new_variables(nc_paths, dataset_codes_dict, version_date, func_dict):
+def multi_calc_new_variables(nc_paths, dataset_codes_dict, version_date, func_dict, max_workers=4):
     """
 
     """
     dates1 = misc.filter_data_paths(nc_paths, '%Y%m%d%H%M%S', return_dates=True)
     dates1['variable'] = dates1.path.apply(lambda x: pathlib.Path(x).stem.split('_wgs84_')[0])
 
+    dill_func_dict = dill.dumps(func_dict)
+
     ## Iterate through files
     # if max_workers <=1:
-    new_paths_list = []
-    for i, g in dates1.groupby('date'):
-        # print(g)
-        p = processing.calc_new_variables(g['path'].tolist(), dataset_codes_dict, version_date, func_dict)
-        new_paths_list.append(p)
+    # new_paths_list = []
+    # for i, g in dates1.groupby('date'):
+    #     # print(g)
+    #     p = processing.calc_new_variables(g['path'].tolist(), dataset_codes_dict, version_date, func_dict)
+    #     new_paths_list.append(p)
 
     # elif max_workers > 1:
-    #     with concurrent.futures.ProcessPoolExecutor(max_workers=max_workers, mp_context=mp.get_context("spawn")) as executor:
-    #         futures = []
-    #         for i, g in dates1.groupby('date'):
-    #             # print(g)
-    #             f = executor.submit(processing.calc_new_variables, g['path'].tolist(), datasets, func_dict)
-    #             futures.append(f)
-    #         runs = concurrent.futures.wait(futures)
+    with concurrent.futures.ProcessPoolExecutor(max_workers=max_workers, mp_context=mp.get_context("spawn")) as executor:
+        futures = []
+        for i, g in dates1.groupby('date'):
+            # print(g)
+            f = executor.submit(processing.calc_new_variables, g['path'].tolist(), dataset_codes_dict, version_date, dill_func_dict)
+            futures.append(f)
+        runs = concurrent.futures.wait(futures)
 
-    #     # process output
-    #     new_paths_list = [r.result() for r in runs[0]]
+    # process output
+    new_paths_list = [r.result() for r in runs[0]]
 
     new_paths1 = []
     for new_path in new_paths_list:
@@ -517,7 +523,7 @@ def multi_mergetime_nc_remove(source_paths, by, max_workers=3):
     with concurrent.futures.ProcessPoolExecutor(max_workers=max_workers, mp_context=mp.get_context("spawn")) as executor:
         futures = []
         for ds_id, paths in source_paths_dict.items():
-            f = executor.submit(data_io.mergetime_nc_remove, paths, by)
+            f = executor.submit(data_io.mergetime_nc_files, paths, by)
             futures.append(f)
         runs = concurrent.futures.wait(futures)
 
@@ -527,15 +533,6 @@ def multi_mergetime_nc_remove(source_paths, by, max_workers=3):
     return new_paths
 
 
-
-
-
-
-
-
-
-
-
 ###############################################
 ### Main class
 
@@ -655,7 +652,7 @@ class Grid(titan.Titan):
 
         variables_set = set()
         for d in dataset_codes:
-            v1 = self._func_dict[d]['variables']
+            v1 = self.func_dict[d]['variables']
             variables_set.update(v1)
 
         variables = list(variables_set)
@@ -676,11 +673,11 @@ class Grid(titan.Titan):
         return new_paths2
 
 
-    def calc_new_variables(self, source_paths):
+    def calc_new_variables(self, source_paths, max_workers=4):
         """
 
         """
-        new_paths2 = multi_calc_new_variables(source_paths, self.dataset_codes_dict, self.max_version_date, self._func_dict)
+        new_paths2 = multi_calc_new_variables(source_paths, self.dataset_codes_dict, self.max_version_date, self.func_dict, max_workers=max_workers)
 
         return new_paths2
 
@@ -741,11 +738,16 @@ class Grid(titan.Titan):
         """
 
         """
+        ## Checks
         misc.diagnostic_check(self.diagnostics, 'status_checks')
 
+        ## Prepare
+        max_size = max_size_gb * 1000000000
+
         key_pattern = tdm.key_patterns[self.system_version]['interim_results'].split('{dataset_id}')[0]
         key = key_pattern.format(run_id=self.run_id)
 
+        ## Query objects
         client = s3_client(self.connection_config)
         obj_list1 = s3.list_objects_s3(client, self.bucket, key)
         obj_list1 = obj_list1[['Key', 'Size']].rename(columns={'Key': 'key', 'Size': 'size'})
@@ -758,29 +760,26 @@ class Grid(titan.Titan):
         obj_list1['station_id'] = stn_keys2
 
         # ds_sizes = obj_list1.groupby('dataset_id')['size'].sum().reset_index()
-        ds_stn_sizes = obj_list1.groupby(['dataset_id', 'station_id'])['size'].sum().reset_index()
-        ds_stn_sizes = ds_stn_sizes.sample(frac=1, ignore_index=True, random_state=1)
+        ds_stn_sizes = obj_list1.groupby(['dataset_id', 'station_id'])['size'].sum().reset_index().sort_values('size', ascending=False)
 
-        # self.ds_sizes = ds_sizes
-        # self.ds_stn_sizes = ds_stn_sizes
-        # self.interim_objects = obj_list1
-
-        max_size = max_size_gb * 1000000000
-
-        ds_stn_sum = ds_stn_sizes['size'].sum()
-        mean_size = ds_stn_sum/len(ds_stn_sizes)
-        group_len = int(max_size//mean_size)
-        df_len = len(ds_stn_sizes)
-
-        grouper = np.repeat(np.arange((df_len//group_len)+1), group_len)
-
-        grp1 = ds_stn_sizes.groupby(grouper[:df_len])
-        ds_stn_groups = [v for k, v in grp1]
+        ## Test to make sure the max size is bigger than the largest file
+        if ds_stn_sizes.iloc[0]['size'] > max_size:
+            raise ValueError('The max_size_gb must be larger than the largest object: {} GB.'.format(round(ds_stn_sizes.iloc[0]['size'] / 1000000000, 2)))
 
+        ## Split the objects
         interim_groups = []
-        for df in ds_stn_groups:
-            ig = pd.merge(df.drop('size', axis=1), obj_list1, on=['dataset_id', 'station_id'])
-            interim_groups.append(ig)
+
+        while True:
+            ds_stn_sizes['cumsum'] = ds_stn_sizes['size'].cumsum()
+            ds_stn_set = ds_stn_sizes[ds_stn_sizes['cumsum'] < max_size]
+
+            if ds_stn_set.empty:
+                break
+            else:
+                ig = pd.merge(ds_stn_set[['dataset_id', 'station_id']], obj_list1, on=['dataset_id', 'station_id'])
+                ds_stn_sizes = ds_stn_sizes[ds_stn_sizes['cumsum'] >= max_size].copy()
+
+                interim_groups.append(ig)
 
         ## diagnostic log
         self.diagnostics['get_interim_results_list'] = {'pass': True}
