@@ -21,11 +21,19 @@ from datetime import date, datetime
 import hashlib
 import xarray as xr
 # from blake3 import blake3
+import zlib
+from time import sleep
+import uuid
+import pathlib
+from pydantic import HttpUrl
+from typing import Optional, List, Any, Union
+import botocore
+from tethysts.utils import s3_client, get_object_s3, create_public_s3_url, url_stream_to_file
 
 ############################################
 ### Parameters
 
-
+local_results_name = '{ds_id}/{stn_id}/{chunk_id}.{version_date}.{chunk_hash}.nc'
 
 ############################################
 ### Functions
@@ -95,7 +103,7 @@ def make_run_date_key(run_date=None):
     return run_date_key
 
 
-def write_json_zstd(data, file_path=None, compress_level=1):
+def write_json_zstd(data, file_path=None, compress_level=1, retries=3):
     """
     Serializer of a python dictionary to json using orjson then compressed using zstandard.
 
@@ -111,19 +119,61 @@ def write_json_zstd(data, file_path=None, compress_level=1):
     bytes object
     """
     json1 = orjson.dumps(data, option=orjson.OPT_NON_STR_KEYS | orjson.OPT_OMIT_MICROSECONDS | orjson.OPT_SERIALIZE_NUMPY)
-    cctx = zstd.ZstdCompressor(level=compress_level)
-    c_obj = cctx.compress(json1)
 
-    if isinstance(file_path, str):
-        with open(file_path, 'wb') as p:
-            p.write(c_obj)
+    # counter = retries
+    # while True:
+    #     try:
+    #         cctx = zstd.ZstdCompressor(level=compress_level, write_checksum=True, write_content_size=True)
+    #         c_obj = cctx.compress(json1)
+
+    #         ## Test
+    #         ddtx = zstd.ZstdDecompressor()
+    #         _ = ddtx.decompress(c_obj)
+    #         break
+
+    #     except Exception as err:
+    #         print(str(err))
+    #         sleep(2)
+    #         counter = counter - 1
+    #         if counter <= 0:
+    #             raise err
+
+    if isinstance(file_path, (str, pathlib.Path)):
+        # source_checksum = zlib.adler32(c_obj)
+
+        # counter = retries
+        # while True:
+        #     try:
+        #         with open(file_path, 'wb') as p:
+        #             p.write(c_obj)
+
+        #         ## Test the write
+        #         with open(file_path, 'rb') as f:
+        #             file_checksum = zlib.adler32(f.read())
+
+        #         if source_checksum == file_checksum:
+        #             break
+        #         else:
+        #             raise ValueError('checksum mismatch...')
+        #     except Exception as err:
+        #         print(str(err))
+        #         sleep(2)
+        #         counter = counter - 1
+        #         if counter <= 0:
+        #             raise err
+
+        with open(file_path, 'wb') as f:
+            cctx = zstd.ZstdCompressor(level=compress_level, write_content_size=True)
+            with cctx.stream_writer(f, size=len(json1)) as compressor:
+                compressor.write(json1)
     else:
+        cctx = zstd.ZstdCompressor(level=compress_level, write_content_size=True)
+        c_obj = cctx.compress(json1)
+
         return c_obj
 
-    return c_obj
 
-
-def write_pkl_zstd(obj, file_path=None, compress_level=1, pkl_protocol=pickle.HIGHEST_PROTOCOL):
+def write_pkl_zstd(obj, file_path=None, compress_level=1, pkl_protocol=pickle.HIGHEST_PROTOCOL, retries=3):
     """
     Serializer using pickle and zstandard. Converts any object that can be pickled to a binary object then compresses it using zstandard. Optionally saves the object to disk. If obj is bytes, then it will only be compressed without pickling.
 
@@ -145,13 +195,56 @@ def write_pkl_zstd(obj, file_path=None, compress_level=1, pkl_protocol=pickle.HI
     else:
         p_obj = pickle.dumps(obj, protocol=pkl_protocol)
 
-    cctx = zstd.ZstdCompressor(level=compress_level)
-    c_obj = cctx.compress(p_obj)
+    # counter = retries
+    # while True:
+    #     try:
+    #         cctx = zstd.ZstdCompressor(level=compress_level, write_checksum=True, write_content_size=True)
+    #         c_obj = cctx.compress(p_obj)
 
-    if isinstance(file_path, str):
-        with open(file_path, 'wb') as p:
-            p.write(c_obj)
+    #         ## Test
+    #         ddtx = zstd.ZstdDecompressor()
+    #         _ = ddtx.decompress(c_obj)
+    #         break
+
+    #     except Exception as err:
+    #         print(str(err))
+    #         sleep(2)
+    #         counter = counter - 1
+    #         if counter <= 0:
+    #             raise err
+
+    if isinstance(file_path, (str, pathlib.Path)):
+        # source_checksum = zlib.adler32(c_obj)
+
+        # counter = retries
+        # while True:
+        #     try:
+        #         with open(file_path, 'wb') as p:
+        #             p.write(c_obj)
+
+        #         ## Test the write
+        #         with open(file_path, 'rb') as f:
+        #             file_checksum = zlib.adler32(f.read())
+
+        #         if source_checksum == file_checksum:
+        #             break
+        #         else:
+        #             raise ValueError('checksum mismatch...')
+        #     except Exception as err:
+        #         print(str(err))
+        #         sleep(2)
+        #         counter = counter - 1
+        #         if counter <= 0:
+        #             raise err
+
+        with open(file_path, 'wb') as f:
+            cctx = zstd.ZstdCompressor(level=compress_level, write_content_size=True)
+            with cctx.stream_writer(f, size=len(p_obj)) as compressor:
+                compressor.write(p_obj)
     else:
+        cctx = zstd.ZstdCompressor(level=compress_level, write_content_size=True)
+        c_obj = cctx.compress(p_obj)
+
         return c_obj
 
 
@@ -450,12 +543,57 @@ def clear_xr_objects():
     return n_cleaned
 
 
+def check_compressor(obj):
+    """
 
+    """
+    cctx = zstd.ZstdCompressor(level=1, write_checksum=True, write_content_size=True)
+    c_obj = cctx.compress(obj)
 
+    ## Test
+    ddtx = zstd.ZstdDecompressor()
+    _ = ddtx.decompress(c_obj)
 
 
+def check_writing(obj, base_path):
+    """
 
+    """
+    id1 = uuid.uuid4().hex
+    file_path = os.path.join(base_path, id1 + '.zst')
+    source_checksum = zlib.adler32(obj)
 
+    with open(file_path, 'wb') as p:
+        p.write(obj)
+
+    ## Test the write
+    with open(file_path, 'rb') as f:
+        file_checksum = zlib.adler32(f.read())
+
+    if source_checksum != file_checksum:
+        raise ValueError('checksum mismatch...')
+
+
+def download_results(chunk: dict, bucket: str, cache: pathlib.Path, s3: botocore.client.BaseClient = None, connection_config: dict = None, public_url: HttpUrl = None):
+    """
+    """
+    chunk_hash = chunk['chunk_hash']
+    version_date = pd.Timestamp(chunk['version_date']).strftime('%Y%m%d%H%M%SZ')
+    results_file_name = local_results_name.format(ds_id=chunk['dataset_id'], stn_id=chunk['station_id'], chunk_id=chunk['chunk_id'], version_date=version_date, chunk_hash=chunk_hash)
+    chunk_path = cache.joinpath(results_file_name)
+    chunk_path.parent.mkdir(parents=True, exist_ok=True)
+
+    if not chunk_path.exists():
+        if public_url is not None:
+            url = create_public_s3_url(public_url, bucket, chunk['key'])
+            _ = url_stream_to_file(url, chunk_path, compression='zstd')
+        else:
+            obj1 = get_object_s3(chunk['key'], bucket, s3, connection_config, public_url)
+            with open(chunk_path, 'wb') as f:
+                f.write(obj1)
+            del obj1
+
+    return {'station_id': chunk['station_id'], 'chunk': chunk_path}
 
 
 
