@@ -10,7 +10,7 @@ import numpy as np
 import xarray as xr
 import pandas as pd
 from tethys_utils import misc, s3, processing
-from tethysts.utils import s3_client, get_object_s3, read_json_zstd, read_pkl_zstd, download_results, create_public_s3_url
+from tethysts.utils import s3_client, get_object_s3, read_json_zstd, read_pkl_zstd, create_public_s3_url, make_run_date_key
 import tethys_data_models as tdm
 from typing import List, Optional, Dict, Union
 from botocore import exceptions as bc_exceptions
@@ -26,6 +26,7 @@ import glob
 import multiprocessing as mp
 import random
 from time import sleep
+import shutil
 
 ##############################################
 ### Parameters
@@ -367,22 +368,21 @@ class Titan(object):
         """
         misc.diagnostic_check(self.diagnostics, 'load_dataset_metadata')
 
-        version_dict, vd_old = s3.process_dataset_versions(self.dataset_list, bucket=self.bucket, connection_config=self.connection_config, public_url=self.public_url, version_data=version_data, system_version=self.system_version)
+        version_dict = s3.process_dataset_versions(self.dataset_list, bucket=self.bucket, connection_config=self.connection_config, public_url=self.public_url, version_data=version_data, system_version=self.system_version)
         max_version_date = max(list([d['version_date'] for d in version_dict.values()]))
 
         # setattr(self, 'processing_code', processing_code)
         setattr(self, 'version_dict', version_dict)
         setattr(self, 'max_version_date', max_version_date)
-        setattr(self, 'old_versions', vd_old)
 
         ## diagnostic log
         self.diagnostics['process_versions'] = {'pass': True}
-        self.diagnostics['attributes'].update({'version_dict': version_dict, 'max_version_date': max_version_date, 'old_versions': vd_old})
+        self.diagnostics['attributes'].update({'version_dict': version_dict, 'max_version_date': max_version_date})
 
         return version_dict
 
 
-    def _save_new_results(self, source_paths, max_workers=4):
+    def _save_new_results(self, source_paths, check_dupes=False, max_workers=4):
         """
 
         """
@@ -462,7 +462,7 @@ class Titan(object):
                 futures = []
                 for chunk in chunks:
                     remote2['chunk'] = chunk
-                    f = executor.submit(download_results, **remote2)
+                    f = executor.submit(misc.download_results, **remote2)
                     futures.append(f)
                 runs = concurrent.futures.wait(futures)
 
@@ -516,7 +516,7 @@ class Titan(object):
         return updated_paths
 
 
-    def _update_results_chunks(self, results_paths, rc_diffs, max_workers=4):
+    def _update_results_chunks(self, results_paths, rc_diffs, old_versions, max_workers=4):
         """
 
         """
@@ -527,10 +527,11 @@ class Titan(object):
         if self._add_old:
             misc.diagnostic_check(self.diagnostics, 'update_conflicting_results')
 
-        r_chunks1 = rc_diffs[rc_diffs['_merge'] == 'conflict']
+        if rc_diffs is not None:
+            r_chunks1 = rc_diffs[rc_diffs['_merge'] == 'conflict']
 
-        if not r_chunks1.empty:
-            misc.diagnostic_check(self.diagnostics, 'update_conflicting_results')
+            if not r_chunks1.empty:
+                misc.diagnostic_check(self.diagnostics, 'update_conflicting_results')
 
         ## Organise the paths
         paths_dict = {}
@@ -544,31 +545,27 @@ class Titan(object):
                 paths_dict[ds_id] = [path]
 
         ## Iterate through datasets
-        with concurrent.futures.ProcessPoolExecutor(max_workers=max_workers, mp_context=mp.get_context("spawn")) as executor:
-            futures = []
+        rc_paths = []
 
-            for ds_id, stn_files in paths_dict.items():
+        for ds_id, stn_files in paths_dict.items():
 
-                if ds_id in self.old_versions:
-                    old_versions = [v['version_date'] for v in self.old_versions[ds_id]]
-                else:
-                    old_versions = []
+            if ds_id in old_versions:
+                old_versions1 = [v['version_date'] for v in old_versions[ds_id]]
+            else:
+                old_versions1 = []
 
-                ## Get the old results chunks file
-                if self.max_version_date in old_versions:
-                    old_rc_data = s3.get_remote_results_chunks(self.bucket, s3=None, connection_config=self.connection_config, public_url=self.public_url, dataset_id=ds_id, version_date=self.max_version_date, system_version=self.system_version)
-                elif add_old and (len(old_versions) > 0):
-                    previous_version_date = old_versions[-1]
-                    old_rc_data = s3.get_remote_results_chunks(self.bucket, s3=None, connection_config=self.connection_config, public_url=self.public_url, dataset_id=ds_id, version_date=previous_version_date, system_version=self.system_version)
-                else:
-                    old_rc_data = None
+            ## Get the old results chunks file
+            if self.max_version_date in old_versions1:
+                old_rc_data = s3.get_remote_results_chunks(self.bucket, dataset_id=ds_id, version_date=self.max_version_date, s3=None, connection_config=self.connection_config, public_url=self.public_url,  system_version=self.system_version)
+            elif add_old and (len(old_versions1) > 0):
+                previous_version_date = old_versions1[-1]
+                old_rc_data = s3.get_remote_results_chunks(self.bucket, dataset_id=ds_id, version_date=previous_version_date, s3=None, connection_config=self.connection_config, public_url=self.public_url,  system_version=self.system_version)
+            else:
+                old_rc_data = None
 
-                f = executor.submit(processing.update_results_chunks, stn_files, self.preprocessed_path, old_rc_data)
-                futures.append(f)
-
-            runs = concurrent.futures.wait(futures)
-
-        rc_paths = [r.result() for r in runs[0]]
+            ## Run through all files
+            rc_path = processing.update_results_chunks(stn_files, self.preprocessed_path, old_rc_data, max_workers)
+            rc_paths.append(rc_path)
 
         self.results_chunks_paths = rc_paths
 
@@ -578,7 +575,7 @@ class Titan(object):
         return rc_paths
 
 
-    def _update_stations(self, results_paths, max_workers=4):
+    def _update_stations(self, results_paths, old_versions, max_workers=4):
         """
 
         """
@@ -607,32 +604,27 @@ class Titan(object):
             rc_path_dict[ds_id] = path
 
         ## Iterate through datasets
-        with concurrent.futures.ProcessPoolExecutor(max_workers=max_workers, mp_context=mp.get_context("spawn")) as executor:
-            futures = []
+        stn_paths = []
 
-            for ds_id, data_paths in paths_dict.items():
-                rc_path = rc_path_dict[ds_id]
+        for ds_id, data_paths in paths_dict.items():
+            rc_path = rc_path_dict[ds_id]
 
-                if ds_id in self.old_versions:
-                    old_versions = [v['version_date'] for v in self.old_versions[ds_id]]
-                else:
-                    old_versions = []
+            if ds_id in old_versions:
+                old_versions1 = [v['version_date'] for v in old_versions[ds_id]]
+            else:
+                old_versions1 = []
 
-                ## Get the old results chunks file
-                if self.max_version_date in old_versions:
-                    old_stns_data = s3.get_remote_stations(self.bucket, s3=None, connection_config=self.connection_config, public_url=self.public_url, dataset_id=ds_id, version_date=self.max_version_date, system_version=self.system_version)
-                elif add_old and (len(old_versions) > 0):
-                    previous_version_date = old_versions[-1]
-                    old_stns_data = s3.get_remote_stations(self.bucket, s3=None, connection_config=self.connection_config, public_url=self.public_url, dataset_id=ds_id, version_date=previous_version_date, system_version=self.system_version)
-                else:
-                    old_stns_data = None
+            ## Get the old results chunks file
+            if self.max_version_date in old_versions1:
+                old_stns_data = s3.get_remote_stations(self.bucket, dataset_id=ds_id, version_date=self.max_version_date, s3=None, connection_config=self.connection_config, public_url=self.public_url,  system_version=self.system_version)
+            elif add_old and (len(old_versions1) > 0):
+                previous_version_date = old_versions1[-1]
+                old_stns_data = s3.get_remote_stations(self.bucket, dataset_id=ds_id, version_date=previous_version_date, s3=None, connection_config=self.connection_config, public_url=self.public_url,  system_version=self.system_version)
+            else:
+                old_stns_data = None
 
-                f = executor.submit(processing.update_stations, data_paths, rc_path, self.preprocessed_path, old_stns_data)
-                futures.append(f)
-
-            runs = concurrent.futures.wait(futures)
-
-        stn_paths = [r.result() for r in runs[0]]
+            stn_path = processing.update_stations(data_paths, rc_path, self.preprocessed_path, old_stns_data, max_workers)
+            stn_paths.append(stn_path)
 
         self.stations_paths = stn_paths
 
@@ -642,7 +634,7 @@ class Titan(object):
         return stn_paths
 
 
-    def _update_versions(self, max_workers=4):
+    def _update_versions(self, old_versions, max_workers=4):
         """
 
         """
@@ -662,16 +654,12 @@ class Titan(object):
                 version_data1 = self.version_dict[dataset_id]
 
                 ## Get the versions
-                if dataset_id in self.old_versions:
-                    rv_key = tdm.utils.key_patterns[self.system_version]['versions'].format(dataset_id=dataset_id)
-
-                    v_obj1 = get_object_s3(rv_key, self.bucket, connection_config=self.connection_config, public_url=self.public_url)
-
-                    old_versions = read_json_zstd(v_obj1)
+                if dataset_id in old_versions:
+                    old_versions1 = old_versions[dataset_id]
                 else:
-                    old_versions = None
+                    old_versions1 = None
 
-                f = executor.submit(processing.update_versions, version_data1, self.preprocessed_path, old_versions)
+                f = executor.submit(processing.update_versions, version_data1, self.preprocessed_path, old_versions1)
                 futures.append(f)
 
             runs = concurrent.futures.wait(futures)
@@ -739,12 +727,12 @@ class Titan(object):
             ## Checks
             misc.diagnostic_check(self.diagnostics, 'update_datasets')
 
-            ds_set = set()
-            for p in results_paths:
-                ds_id = os.path.split(p)[1].split('_')[0]
-                ds_set.add(ds_id)
+            # ds_set = set()
+            # for p in results_paths:
+            #     ds_id = os.path.split(p)[1].split('_')[0]
+            #     ds_set.add(ds_id)
 
-            if not (len(rc_paths) == len(stns_paths) == len(ds_paths) == len(ds_set) == len(v_paths)):
+            if not (len(rc_paths) == len(stns_paths) == len(ds_paths) == len(v_paths)):
                 raise ValueError('The number of auxillary files do not match up.')
 
             if len(dss_path) != 1:
@@ -786,7 +774,10 @@ class Titan(object):
         ## Remove all files in the results path
         path1 = pathlib.Path(self.preprocessed_path)
         for path in path1.glob('*'):
-            os.remove(path)
+            if path.is_dir():
+                shutil.rmtree(path)
+            else:
+                os.remove(path)
 
         return resp
 
@@ -804,29 +795,55 @@ class Titan(object):
 
         if len(results_paths1) > 0:
 
+            results_paths2 = list(set(results_paths1))
+
+            ## Check for duplicate chunk_ids for the same dataset and station
+            file_name = os.path.split(results_paths2[0])[1]
+
+            if len(file_name) > 50:
+                seen = set()
+                dupes = []
+
+                for x in results_paths2:
+                    p = x[:-40]
+                    if p in seen:
+                        dupes.append(p)
+                    else:
+                        seen.add(p)
+
+                if dupes:
+                    raise ValueError('Duplicate dataset/station/chunk(s) found: ' + str(dupes))
+
             ## Determine the chunks diffs
             if isinstance(self.public_url, str):
                 remote1 = {'bucket': self.bucket, 'public_url': self.public_url, 'version': self.system_version}
             else:
                 remote1 = {'bucket': self.bucket, 'connection_config': self.connection_config, 'version': self.system_version}
 
-            results_paths_set = set(results_paths1)
-
-            rc_diffs = self._determine_results_chunks_diffs(list(results_paths_set), remote1, add_old=self._add_old, max_workers=max_workers)
+            rc_diffs = self._determine_results_chunks_diffs(results_paths2, remote1, add_old=self._add_old, max_workers=max_workers)
 
             print('-- Check and update conflicting results chunks...')
-            results_paths_new = self._update_conflicting_results(results_paths, rc_diffs, threads=threads, max_workers=max_workers)
+            results_paths_new = self._update_conflicting_results(results_paths2, rc_diffs, threads=threads, max_workers=max_workers)
 
             if len(results_paths_new) > 0:
 
+                ## Get old versions
+                ds_ids = set()
+                for path in list(set(results_paths_new)):
+                    path1 = os.path.split(path)[1]
+                    ds_id, version_date1, stn_id, chunk_id, _, _ = path1.split('_')
+                    ds_ids.add(ds_id)
+
+                old_versions = s3.get_versions(list(ds_ids), self.bucket, connection_config=self.connection_config, public_url=self.public_url, system_version=self.system_version)
+
                 print('-- Updating the results chunks...')
-                rc_paths = self._update_results_chunks(results_paths_new, rc_diffs, max_workers)
+                rc_paths = self._update_results_chunks(results_paths_new, rc_diffs, old_versions, max_workers)
 
                 print('-- Updating the stations...')
-                stns_paths = self._update_stations(results_paths_new, max_workers)
+                stns_paths = self._update_stations(results_paths_new, old_versions, max_workers)
 
                 print('-- Updating the versions...')
-                versions_paths = self._update_versions(max_workers)
+                versions_paths = self._update_versions(old_versions, max_workers)
 
                 print('-- Updating the datasets...')
                 ds_paths = self._update_datasets()
@@ -847,6 +864,70 @@ class Titan(object):
         return resp
 
 
+    def _reset_auxillary_objects(self, dataset_id, threads=80, max_workers=4):
+        """
+
+        """
+        if dataset_id not in self.datasets:
+            raise ValueError('{} dataset_id not available.'.format(dataset_id))
+
+        ## Create dir for existing data
+        existing_data_path = os.path.join(self.preprocessed_path, 'existing_data')
+        os.makedirs(existing_data_path, exist_ok=True)
+
+        ## Determine all of the object to download
+        vd = self.max_version_date
+        vd_key = make_run_date_key(vd)
+
+        key_prefix = tdm.utils.key_patterns[self.system_version]['results'].split('{station_id}')[0].format(dataset_id=dataset_id, version_date=vd_key)
+
+        client = s3_client(self.connection_config, threads)
+
+        obj_list1 = s3.list_objects_s3(client, self.bucket, key_prefix)
+        obj_list1 = obj_list1[obj_list1.Key.str.contains('results')].copy()
+
+        ## Download data
+        print('-- Downloading all results chunks ({})...'.format(len(obj_list1)))
+        with concurrent.futures.ThreadPoolExecutor(max_workers=threads) as executor:
+            futures = []
+
+            for key in obj_list1.Key.tolist():
+                f = executor.submit(s3.download_results, key, existing_data_path, self.bucket, client, public_url=self.public_url)
+                futures.append(f)
+
+            runs = concurrent.futures.wait(futures)
+
+        new_paths = [r.result() for r in runs[0]]
+
+        ## Keep the old_version for later
+        old_versions = copy.deepcopy(self.old_versions)
+        self.old_versions = {}
+
+        ## Apply fake completion
+        self.diagnostics['update_conflicting_results'] = {'pass': True}
+        self.diagnostics['save_new_results'] = {'pass': True}
+
+        ## Update aux files
+        print('-- Updating the results chunks...')
+        rc_paths = self._update_results_chunks(new_paths, None, max_workers)
+
+        print('-- Updating the stations...')
+        stns_paths = self._update_stations(new_paths, max_workers)
+
+        print('-- Updating the versions...')
+        self.old_versions = old_versions
+        versions_paths = self._update_versions(1)
+
+        print('-- Updating the datasets...')
+        ds_paths = self._update_datasets()
+
+        print('-- Uploading final objects...')
+        resp = self._upload_final_objects(threads=threads)
+
+        return resp
+
+
+
 
 
 
